/*  bin2rom.c : conversion utility. 
    Copyright (C) 1999  Ludovic LANGE <ludovic.lange@free.fr>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/mman.h>

#define BLOCKSIZE 65535

#define min(a, b) (a<b)?(a):(b)

int bin2rom( char * inputName, char * outputName, long start, long boot )
{
    int input;
    int output;
    char * inStart = NULL;

    char * inIndex;
    struct stat inBuf;
    int status;
    size_t outSize;
    long nullData;
    char header[]="B000FF\n";
    unsigned int checksum;
    unsigned char *pCheckSum;
    unsigned int ptr;
    unsigned int roundedSize;
    unsigned int blockSize;
    
    int reste;
    
    printf("bin2rom : input filename = %s, output filename = %s, start = 0x%lx, boot=0x%lx\n", inputName, outputName, start, boot );

    input = open( inputName, O_RDONLY );
    if( input == -1 )
    {
	perror("bin2rom : input: ");
	exit( errno );
    }
    

    printf("bin2rom : input filedes = %d\n", input );
    
//Let's compute the input file size
//=================================

    status = fstat( input, &inBuf );
    
    if( status != 0 )
    {
	perror("bin2rom : fstat in: ");
	close( input );
	exit( errno );
    }
    
    printf("bin2rom : input filesize = %d (bytes)\n", (int)inBuf.st_size );

    roundedSize = inBuf.st_size;
    if( (roundedSize%4) != 0 )
    {
	roundedSize = (inBuf.st_size - (inBuf.st_size%4) ) + 4;
    }
    
//Map the input file in memory
//============================    
    
    inStart = mmap( (void *)0, (size_t) roundedSize, PROT_READ, MAP_SHARED ,
	input, (off_t) 0 );

    printf("Input Mem : start=%p length=%X (end=%p)\n",inStart, roundedSize, inStart+roundedSize);

    if( inStart == MAP_FAILED )
    {
	perror("bin2rom : mmap out: ");
	close( input );
	exit( errno );
    }
    
    outSize = roundedSize + sizeof( header )-1 + 8 + 12 + 12;
    
    printf("bin2rom: Image length = %X\n", outSize );

//Create and fill output file
//===========================

    output = open( outputName, O_CREAT | O_RDWR | O_TRUNC, S_IRUSR | S_IWUSR );
    if( output == -1 )
    {
	perror("bin2rom : output: ");
	exit( errno );
    }

#ifdef MAP_OUTPUT
    
    status = lseek( output, outSize, SEEK_SET );

    if( status  == -1 )
    {
	perror("bin2rom : lseek out: ");
	close( input );
	close( output );
	exit( errno );
    }

    write( output, inIndex, 1);

    close( output );

    output = open( outputName, O_RDWR , S_IRUSR | S_IWUSR );
    if( output == -1 )
    {
	perror("bin2rom : output: ");
	exit( errno );
    }
#endif

    printf("bin2rom : input filedes = %d, output filedes = %d\n", input, output );
    
//Map the output file in memory
//=============================

#ifdef MAP_OUTPUT
    outStart = mmap( (void *)0, outSize, PROT_WRITE , MAP_SHARED ,
	output, (off_t) 0 );
    outIndex = outStart;

    printf("Output Mem : start=%X length=%X (end=%X)\n",outStart, outSize, outStart+outSize);
    if( outStart == MAP_FAILED )
    {
	perror("bin2rom : mmap out: ");
	close( input );
	close( output );
	exit( errno );
    }
#endif    
    
//Write Header + start + size
//===========================	

#ifdef MAP_OUTPUT
    memcpy( outIndex, header, sizeof(header)-1 );
    outIndex += sizeof( header )-1;

    memcpy( outIndex, &start, 4 );
    outIndex +=4;

    memcpy( outIndex, &( roundedSize ), 4 );
    outIndex +=4;

#else

    write( output, header, sizeof( header )-1 );
	
    write( output, &start, 4 );
    
    write( output, &roundedSize, 4 );
#endif

//Compute the checksum
//====================
    checksum = 0;
    pCheckSum = (unsigned char *) inStart;
    for( ptr=0; ptr< roundedSize; ptr++ )
    {
	checksum += *pCheckSum++;
    }
	
//Unmap input block
//=================

    status = munmap( inStart, (size_t) roundedSize );
    if( status != 0 )
    {
	perror("bin2rom : munmap in: ");
	close( input );
	close( output );
	exit( errno );
    }

    close( input );

    input = open( inputName, O_RDONLY );
    if( input == -1 )
    {
	perror("bin2rom : input: ");
	exit( errno );
    }
    
//Write the first file header
//===========================

#ifdef MAP_OUTPUT
    memcpy( outIndex, &start, 4 );
    outIndex += 4;
    	
    memcpy( outIndex, &( roundedSize ), 4 );
    outIndex += 4;
	
    memcpy( outIndex, &checksum, 4 );	//Checksum !
    outIndex += 4;
#else
    write( output, &start, 4 );
    
    write( output, &roundedSize, 4);

    write( output, &checksum, 4);

#endif

//Send the block to output.
//=========================

    inStart = (char *) malloc( BLOCKSIZE );
    
    if( inStart == NULL )
    {
	perror("bin2rom : malloc: ");
	close( input );
	close( output );
	exit( errno );
    }

    reste = roundedSize;
    inIndex = inStart;
    
    printf("\n");
    
    while( reste > 0 )
    {
	blockSize = min( BLOCKSIZE, reste );
	
	read( input, inStart, blockSize );

#ifdef MAP_OUTPUT
        memcpy( outIndex, inIndex, blockSize );    
        outIndex += blockSize;
	inIndex += blockSize;

#else

	write( output, inStart, blockSize );
#endif	
	reste -= blockSize;
	printf(".");
    }
    
    printf("\n");
//Write the last header (boot address)
//====================================

    nullData = 0x0;

#ifdef MAP_OUTPUT
    memcpy( outIndex, &nullData, 4 );
    outIndex += 4;
	
    memcpy( outIndex, &boot, 4 );
    outIndex += 4;
	
    memcpy( outIndex, &nullData, 4 );	//Checksum !
    outIndex += 4;

#else

    write( output, &nullData, 4 );

    write( output, &boot, 4);

    write( output, &nullData, 4);

#endif


//Synchronize the mapping
//=======================
#ifdef MAP_OUTPUT
    status = msync( outStart, (size_t) outSize, MS_SYNC );

    if( status != 0 )
    {
	perror("bin2rom : msync out: ");
	close( input );
	close( output );
	exit( errno );
    }
#endif    
//Synchronize the filesystem
//=============================
    status = fsync( output );

    if( status != 0 )
    {
	perror("bin2rom : fsync out: ");
	close( input );
	close( output );
	exit( errno );
    }

//Unmap output file
//=================
#ifdef MAP_OUTPUT
    status = munmap( outStart, outSize );
    
    if( status != 0 )
    {
	perror("bin2rom : munmap out: ");
	close( input );
	close( output );
	exit( errno );
    }
#endif    
    close( output );
    close( input );

    return 0;

}
    
int main( int argc, char * argv[] )
{
    long start;
    long boot;
    
    if( argc < 5 )
    {
	printf("bin2rom: <input file> <output file> <start address> <boot address>\n");
	exit( -1 );
    }
    
    sscanf( argv[3], "%lx", &start );
    sscanf( argv[4], "%lx", &boot );
        
    bin2rom( argv[1], argv[2], start, boot );
    
    return 0;
}